--!A cross-platform build utility based on Lua
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-- Copyright (C) 2015-present, Xmake Open Source Community.
--
-- @author      wuzhenqing
-- @file        xmake.lua
--

-- Huawei Ascend C / bisheng toolchain
--
-- @code
-- target("example")
--     set_toolchains("ascendc")
--     add_ascnpuarchs("dav-2201")
-- @endcode
--
toolchain("ascendc")

    -- set homepage
    set_homepage("https://www.hiascend.com/")
    set_description("Huawei Ascend C (bisheng compiler driver)")

    -- set toolset
    set_toolset("asc",  "bisheng")
    set_toolset("aicpu", "bisheng")
    set_toolset("ascld", "bisheng")
    set_toolset("ascsh", "bisheng")
    set_toolset("ar",   "llvm-ar", "ar")

    -- check toolchain
    on_check("check")

    -- on load
    on_load("load")
