/*!The Treasure Box Library
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Copyright (C) 2009-present, TBOOX Open Source Group.
 *
 * @author      ruki
 * @file        page.c
 * @ingroup     platform
 *
 */

/* //////////////////////////////////////////////////////////////////////////////////////
 * trace
 */
#define TB_TRACE_MODULE_NAME            "page"
#define TB_TRACE_MODULE_DEBUG           (0)

/* //////////////////////////////////////////////////////////////////////////////////////
 * implementation
 */
#include "page.h"
#ifdef TB_CONFIG_OS_WINDOWS
#   include "windows/page.c"
#elif defined(TB_CONFIG_POSIX_HAVE_GETPAGESIZE) || \
		defined(TB_CONFIG_POSIX_HAVE_SYSCONF)
#   include "posix/page.c"
#else
tb_bool_t tb_page_init()
{
    return tb_true;
}
tb_void_t tb_page_exit()
{
}
tb_size_t tb_page_size()
{
    // default: 4KB
    return 4096;
}
#endif
